package com.example;

import com.google.common.base.Stopwatch;
import com.google.zxing.BinaryBitmap;
import com.google.zxing.ChecksumException;
import com.google.zxing.FormatException;
import com.google.zxing.NotFoundException;
import com.google.zxing.Result;
import com.google.zxing.client.j2se.BufferedImageLuminanceSource;
import com.google.zxing.common.HybridBinarizer;
import com.google.zxing.qrcode.QRCodeReader;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.pdfbox.Loader;
import org.apache.pdfbox.rendering.ImageType;
import org.apache.pdfbox.rendering.PDFRenderer;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Optional;

@Slf4j
public class ZxingExam {
    public static QRCodeReader qrCodeReader = new QRCodeReader();

    @SneakyThrows
    public static void main(String[] args) {
        var renderDpi = 72 * 6;
        Optional<String> qrCode;

        var imageFile = Path.of("c:/test", "bbbb.png").toFile();
        var pdfFile = Path.of("c:/test", "bbbb.pdf").toFile();

        for (var quadrant = 0; quadrant <= 4; quadrant++) {
            qrCode = readQrCodeByImage(imageFile, quadrant);
            logQrCode(qrCode, "Image", quadrant);

            qrCode = readQrCodeByPdf(pdfFile, quadrant, renderDpi);
            logQrCode(qrCode, "Pdf", quadrant);
        }
    }

    @SuppressWarnings("OptionalUsedAsFieldOrParameterType")
    private static void logQrCode(Optional<String> qrCode, String source, int quadrant) {
        qrCode.ifPresent(s -> log.info("Success. source:{}, quadrant:{}, qrCode:{}", source, quadrant, s));
    }

    private static Optional<String> readQrCodeByImage(File imageFile, int quadrant) {
        BufferedImage bufferedImage;
        try {
            bufferedImage = getBufferedImageForImage(imageFile);
        } catch (IOException e) {
            return Optional.empty();
        }
        return readQrCodeByBufferedImage(bufferedImage, quadrant);
    }

    private static BufferedImage getQuadrantBufferedImage(BufferedImage bufferedImage, int quadrant) {
        var stopWatch = Stopwatch.createStarted();
        int w = bufferedImage.getWidth();
        int h = bufferedImage.getHeight();
        int subW = w / 2;
        int subH = h / 2;
        int subX = 0;
        int subY = 0;
        switch (quadrant) {
            case 1:
                subX = w / 2;
                break;
            case 2:
                break;
            case 3:
                subY = h / 2;
                break;
            case 4:
                subX = w / 2;
                subY = h / 2;
                break;
        }
        var subBufferedImage = bufferedImage.getSubimage(subX, subY, subW, subH);
        log.debug("make subImage. quadrant:{}, Elapsed:{}", quadrant, stopWatch.stop());
        return subBufferedImage;
    }

    private static Optional<String> readQrCodeByPdf(File pdfFile, Integer quadrant, int renderDpi) {
        BufferedImage bufferedImage;
        try {
            bufferedImage = getBufferedImageForPdf(pdfFile, renderDpi);
        } catch (IOException e) {
            return Optional.empty();
        }
        return readQrCodeByBufferedImage(bufferedImage, quadrant);
    }

    private static Optional<String> readQrCodeByBufferedImage(BufferedImage bufferedImage, int quadrant) {
        if (1 <= quadrant && quadrant <= 4) {
            bufferedImage = getQuadrantBufferedImage(bufferedImage, quadrant);
        }

        var bitMap = new BinaryBitmap(new HybridBinarizer(new BufferedImageLuminanceSource(bufferedImage)));
        var stopWatch = Stopwatch.createStarted();
        Result qrCodeResult;
        try {
            qrCodeResult = qrCodeReader.decode(bitMap);
        } catch (NotFoundException | ChecksumException | FormatException e) {
            log.debug("NG Quadrant:{}, Elapsed:{}", quadrant, stopWatch.stop());
            return Optional.empty();
        }
        log.debug("OK Quadrant:{}, Elapsed:{}", quadrant, stopWatch.stop());
        return Optional.of(qrCodeResult.getText());
    }

    private static BufferedImage getBufferedImageForImage(File imageFile) throws IOException {
        return ImageIO.read(imageFile);
    }

    private static BufferedImage getBufferedImageForPdf(File pdfFile, int renderDpi) throws IOException {
        var stopWatch = Stopwatch.createStarted();
        var pdDocument = Loader.loadPDF(pdfFile);
        var pdfRenderer = new PDFRenderer(pdDocument);
        var bufferedImage = pdfRenderer.renderImageWithDPI(0, renderDpi, ImageType.GRAY);
        log.debug("Convert PDF. renderDpi:{}, Elapsed:{}", renderDpi, stopWatch.stop());
        return bufferedImage;
    }
}
